# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.model import ModelSQL, fields
from trytond.modules.company.model import CompanyValueMixin
from trytond.modules.product.product import TemplateFunction
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Eval
from trytond.transaction import Transaction

cost_price_warehouse = fields.Boolean(
    "Cost Price per Warehouse",
    help="Compute product cost price for each warehouse.")


class Configuration(metaclass=PoolMeta):
    __name__ = 'product.configuration'
    cost_price_warehouse = fields.MultiValue(cost_price_warehouse)

    @classmethod
    def default_cost_price_warehouse(cls, **pattern):
        return cls.multivalue_model(
            'cost_price_warehouse').default_cost_price_warehouse()


class ConfigurationCostPriceWarehouse(ModelSQL, CompanyValueMixin):
    "Product Configuration Cost Price per Warehouse"
    __name__ = 'product.configuration.cost_price_warehouse'
    cost_price_warehouse = cost_price_warehouse

    @classmethod
    def default_cost_price_warehouse(cls):
        return False


class Product(metaclass=PoolMeta):
    __name__ = 'product.product'

    @classmethod
    def __setup__(cls):
        super().__setup__()
        readonly = (
            Eval('context', {}).get('cost_price_warehouse', False)
            & ~Eval('context', {}).get('warehouse'))
        if cls.cost_price.states['readonly']:
            cls.cost_price.states['readonly'] |= readonly
        else:
            cls.cost_price.states['readonly'] = readonly

    def multivalue_records(self, field):
        Value = self.multivalue_model(field)
        records = super().multivalue_records(field)
        if issubclass(Value, CostPrice):
            # Sort to get record with empty warehouse at the end
            records = sorted(records, key=lambda r: r.warehouse is None)
        return records

    def get_multivalue(self, name, **pattern):
        pool = Pool()
        Company = pool.get('company.company')
        context = Transaction().context
        if not isinstance(self._fields[name], TemplateFunction):
            Value = self.multivalue_model(name)
            if issubclass(Value, CostPrice) and context.get('company'):
                company = Company(context['company'])
                if company.cost_price_warehouse:
                    pattern.setdefault('warehouse', context.get('warehouse'))
        return super().get_multivalue(name, **pattern)

    def set_multivalue(self, name, value, save=True, **pattern):
        pool = Pool()
        Company = pool.get('company.company')
        context = Transaction().context
        Value = self.multivalue_model(name)
        if issubclass(Value, CostPrice) and context.get('company'):
            company = Company(context['company'])
            if company.cost_price_warehouse:
                pattern.setdefault('warehouse', context.get('warehouse'))
            if not value and not pattern.get('warehouse'):
                return []
        return super().set_multivalue(name, value, save=save, **pattern)

    @classmethod
    def _domain_moves_cost(cls):
        pool = Pool()
        Company = pool.get('company.company')
        context = Transaction().context
        domain = super()._domain_moves_cost()
        if context.get('company'):
            company = Company(context['company'])
            if company.cost_price_warehouse:
                warehouse = context.get('warehouse')
                domain = [
                    domain,
                    ['OR',
                        ('from_location', 'child_of', warehouse, 'parent'),
                        ('from_location.cost_warehouse', '=', warehouse),
                        ('to_location', 'child_of', warehouse, 'parent'),
                        ('to_location.cost_warehouse', '=', warehouse),
                        ],
                    ]
        return domain

    @classmethod
    def _domain_in_moves_cost(cls):
        pool = Pool()
        Company = pool.get('company.company')
        context = Transaction().context
        domain = super()._domain_in_moves_cost()
        if context.get('company'):
            company = Company(context['company'])
            if company.cost_price_warehouse:
                warehouse = context.get('warehouse')
                domain = ['OR',
                    domain,
                    [
                        ('from_location.type', '=', 'storage'),
                        ('to_location.type', '=', 'storage'),
                        ('from_location', 'not child_of', warehouse, 'parent'),
                        ['OR',
                            ('from_location.cost_warehouse', '!=', warehouse),
                            ('from_location.cost_warehouse', '=', None),
                            ],
                        ['OR',
                            ('to_location', 'child_of', warehouse, 'parent'),
                            ('to_location.cost_warehouse', '=', warehouse),
                            ],
                        ]
                    ]
        return domain

    @classmethod
    def _domain_out_moves_cost(cls):
        pool = Pool()
        Company = pool.get('company.company')
        context = Transaction().context
        domain = super()._domain_out_moves_cost()
        if context.get('company'):
            company = Company(context['company'])
            if company.cost_price_warehouse:
                warehouse = context.get('warehouse')
                domain = ['OR',
                    domain,
                    [
                        ('from_location.type', '=', 'storage'),
                        ('to_location.type', '=', 'storage'),
                        ('to_location', 'not child_of', warehouse, 'parent'),
                        ['OR',
                            ('to_location.cost_warehouse', '!=', warehouse),
                            ('to_location.cost_warehouse', '=', None),
                            ],
                        ['OR',
                            ('from_location', 'child_of', warehouse, 'parent'),
                            ('from_location.cost_warehouse', '=', warehouse),
                            ],
                        ]
                    ]
        return domain

    @classmethod
    def _domain_storage_quantity(cls):
        pool = Pool()
        Company = pool.get('company.company')
        context = Transaction().context
        domain = super()._domain_storage_quantity()
        if context.get('company'):
            company = Company(context['company'])
            if company.cost_price_warehouse:
                warehouse = context.get('warehouse')
                domain = [
                    domain,
                    ['OR',
                        ('parent', 'child_of', warehouse),
                        ('cost_warehouse', '=', warehouse),
                        ],
                    ]
        return domain

    @classmethod
    def default_cost_price(cls, **pattern):
        context = Transaction().context
        default = super().default_cost_price()
        if (context.get('cost_price_warehouse')
                and not pattern.get('warehouse', context.get('warehouse'))):
            default = None
        return default


class CostPrice(metaclass=PoolMeta):
    __name__ = 'product.cost_price'
    warehouse = fields.Many2One(
        'stock.location', "Warehouse", ondelete='RESTRICT',
        domain=[
            ('type', '=', 'warehouse'),
            ],
        states={
            'required': Eval('warehouse_required', False),
            })
    warehouse_required = fields.Function(fields.Boolean(
            "Warehouse Required"), 'get_warehouse_required')

    @classmethod
    def get_warehouse_required(cls, records, name):
        return {r.id: r.company.cost_price_warehouse for r in records}


class CostPriceRevision(metaclass=PoolMeta):
    __name__ = 'product.cost_price.revision'
    warehouse = fields.Many2One(
        'stock.location', "Warehouse", ondelete='RESTRICT',
        domain=[
            ('type', '=', 'warehouse'),
            ],
        states={
            'required': Eval('warehouse_required', False),
            })
    warehouse_required = fields.Function(fields.Boolean(
            "Warehouse Required"), 'get_warehouse_required')

    @classmethod
    def default_warehouse(cls):
        pool = Pool()
        Company = pool.get('company.company')
        company_id = cls.default_company()
        if company_id is not None and company_id >= 0:
            company = Company(company_id)
            if company.cost_price_warehouse:
                return Transaction().context.get('warehouse')

    @classmethod
    def get_warehouse_required(cls, records, name):
        return {r.id: r.company.cost_price_warehouse for r in records}

    @classmethod
    def _get_for_product_domain(cls):
        pool = Pool()
        Company = pool.get('company.company')
        domain = super()._get_for_product_domain()
        context = Transaction().context
        if context.get('company'):
            company = Company(context['company'])
            if company.cost_price_warehouse:
                domain = [
                    domain,
                    ('warehouse', '=', context.get('warehouse')),
                    ]
        return domain


class ModifyCostPrice(metaclass=PoolMeta):
    __name__ = 'product.modify_cost_price'

    def get_revision(self, Revision):
        revision = super().get_revision(Revision)
        if revision.company.cost_price_warehouse:
            revision.warehouse = Transaction().context.get('warehouse')
        return revision


class ProductCostHistory(metaclass=PoolMeta):
    __name__ = 'product.product.cost_history'

    @classmethod
    def _non_moves_clause(cls, history_table, user):
        clause = super()._non_moves_clause(history_table, user)
        if user.company and user.company.cost_price_warehouse:
            warehouse_id = user.warehouse.id if user.warehouse else None
            clause &= history_table.warehouse == warehouse_id
        return clause



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.modules.company.tests import CompanyTestMixin
from trytond.tests.test_tryton import ModuleTestCase


class ProductCostWarehouseTestCase(CompanyTestMixin, ModuleTestCase):
    'Test Product Cost Warehouse module'
    module = 'product_cost_warehouse'
    extras = ['account_invoice_stock', 'product_cost_history']


del ModuleTestCase



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.model import ModelSQL, ModelView, dualmethod, fields
from trytond.pool import PoolMeta
from trytond.transaction import Transaction


class Cron(metaclass=PoolMeta):
    __name__ = 'ir.cron'
    warehouses = fields.Many2Many(
        'ir.cron-stock.warehouse', 'cron', 'warehouse',
        "Warehouses",
        domain=[
            ('type', '=', 'warehouse'),
            ],
        help="Warehouses registered for this cron.")

    @dualmethod
    @ModelView.button
    def run_once(cls, crons):
        for cron in crons:
            if not cron.warehouses:
                super().run_once([cron])
            else:
                for warehouse in cron.warehouses:
                    with Transaction().set_context(warehouse=warehouse.id):
                        super(Cron, cls).run_once([cron])


class CronWarehouse(ModelSQL):
    "Cron - Warehouse"
    __name__ = 'ir.cron-stock.warehouse'
    cron = fields.Many2One(
        'ir.cron', "Cron", ondelete='CASCADE', required=True)
    warehouse = fields.Many2One(
        'stock.location', "Warehouse", ondelete='CASCADE', required=True,
        domain=[
            ('type', '=', 'warehouse'),
            ])



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.pool import Pool

from . import account, company, ir, product, res, stock

__all__ = ['register']


def register():
    Pool.register(
        ir.Cron,
        ir.CronWarehouse,
        res.User,
        company.Company,
        product.Configuration,
        product.ConfigurationCostPriceWarehouse,
        product.Product,
        product.CostPrice,
        product.CostPriceRevision,
        stock.Configuration,
        stock.ConfigurationLocation,
        stock.Location,
        stock.Move,
        stock.ShipmentInternal,
        module='product_cost_warehouse', type_='model')
    Pool.register(
        account.InvoiceLine,
        module='product_cost_warehouse', type_='model',
        depends=['account_invoice_stock'])
    Pool.register(
        product.ProductCostHistory,
        module='product_cost_warehouse', type_='model',
        depends=['product_cost_history'])
    Pool.register(
        product.ModifyCostPrice,
        module='product_cost_warehouse', type_='wizard')
    Pool.register(
        module='product_cost_warehouse', type_='report')



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.pool import PoolMeta


class User(metaclass=PoolMeta):
    __name__ = 'res.user'

    @classmethod
    def _get_preferences(cls, user, context_only=False):
        preferences = super()._get_preferences(user, context_only=context_only)
        if user.company:
            preferences['cost_price_warehouse'] = (
                user.company.cost_price_warehouse)
        return preferences



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import copy

from trytond.model import fields
from trytond.pool import Pool, PoolMeta

from .product import cost_price_warehouse


class Company(metaclass=PoolMeta):
    __name__ = 'company.company'

    cost_price_warehouse = fields.Function(
        copy.deepcopy(cost_price_warehouse), 'get_cost_price_warehouse')

    def get_cost_price_warehouse(self, name, **pattern):
        pool = Pool()
        Configuration = pool.get('product.configuration')
        config = Configuration(1)
        pattern['company'] = self.id
        return config.get_multivalue('cost_price_warehouse', **pattern)



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.pool import PoolMeta
from trytond.pyson import Eval


class InvoiceLine(metaclass=PoolMeta):
    __name__ = 'account.invoice.line'

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls.product.context['warehouse'] = Eval('warehouse', -1)
        cls.product.depends.add('warehouse')



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from decimal import Decimal

from trytond.i18n import gettext
from trytond.model import fields
from trytond.modules.product import round_price
from trytond.modules.stock.exceptions import MoveValidationError
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Bool, Eval
from trytond.transaction import Transaction


class Configuration(metaclass=PoolMeta):
    __name__ = 'stock.configuration'

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls.shipment_internal_transit.domain = [
            cls.shipment_internal_transit.domain,
            ('cost_warehouse', '=', None),
            ]


class ConfigurationLocation(metaclass=PoolMeta):
    __name__ = 'stock.configuration.location'

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls.shipment_internal_transit.domain = [
            cls.shipment_internal_transit.domain,
            ('cost_warehouse', '=', None),
            ]


class Location(metaclass=PoolMeta):
    __name__ = 'stock.location'

    cost_warehouse = fields.Many2One(
        'stock.location', "Cost Warehouse",
        domain=[
            ('type', '=', 'warehouse'),
            ],
        states={
            'invisible': (
                (Eval('type') != 'storage')
                | Bool(Eval('warehouse'))
                | (Eval('id', -1) < 0)),
            })


class Move(metaclass=PoolMeta):
    __name__ = 'stock.move'

    @property
    def from_warehouse(self):
        return super().from_warehouse or self.from_location.cost_warehouse

    @property
    def to_warehouse(self):
        return super().to_warehouse or self.to_location.cost_warehouse

    @fields.depends('company', 'from_location', 'to_location')
    def on_change_with_unit_price_required(self, name=None):
        required = super().on_change_with_unit_price_required(name=name)
        if (self.company and self.company.cost_price_warehouse
                and self.from_location and self.to_location
                and self.from_warehouse != self.to_warehouse):
            required = True
        return required

    @fields.depends('company', 'from_location', 'to_location')
    def on_change_with_cost_price_required(self, name=None):
        required = super().on_change_with_cost_price_required(name=name)
        if (self.company and self.company.cost_price_warehouse
                and self.from_location and self.to_location
                and self.from_warehouse != self.to_warehouse):
            required = True
        return required

    @classmethod
    def get_unit_price_company(cls, moves, name):
        pool = Pool()
        ShipmentInternal = pool.get('stock.shipment.internal')
        Uom = pool.get('product.uom')
        prices = super().get_unit_price_company(moves, name)
        for move in moves:
            if (move.company.cost_price_warehouse
                    and move.from_warehouse != move.to_warehouse
                    and move.to_warehouse
                    and isinstance(move.shipment, ShipmentInternal)):
                cost = total_qty = 0
                for outgoing_move in move.shipment.outgoing_moves:
                    if outgoing_move.product == move.product:
                        qty = Uom.compute_qty(
                            outgoing_move.unit, outgoing_move.quantity,
                            move.product.default_uom)
                        qty = Decimal(str(qty))
                        cost += qty * outgoing_move.cost_price
                        total_qty += qty
                if cost and total_qty:
                    cost_price = round_price(cost / total_qty)
                    prices[move.id] = cost_price
        return prices

    def get_cost_price(self, product_cost_price=None):
        pool = Pool()
        ShipmentInternal = pool.get('stock.shipment.internal')
        cost_price = super().get_cost_price(
            product_cost_price=product_cost_price)
        if (self.company.cost_price_warehouse
                and self.from_warehouse != self.to_warehouse
                and self.to_warehouse
                and isinstance(self.shipment, ShipmentInternal)):
            cost_price = self.unit_price_company
        return cost_price

    @classmethod
    def validate(cls, moves):
        pool = Pool()
        Configuration = pool.get('stock.configuration')
        super().validate(moves)
        config = Configuration(1)
        transit_locations = {}
        for move in moves:
            if move.state in {'staging', 'draft'}:
                continue
            company = move.company
            if company not in transit_locations:
                transit_location = config.get_multivalue(
                    'shipment_internal_transit', company=company)
                transit_locations[company] = transit_location
            else:
                transit_location = transit_locations[company]
            if (company.cost_price_warehouse
                    and move.from_location.type == 'storage'
                    and move.from_location != transit_location
                    and move.to_location.type == 'storage'
                    and move.to_location != transit_location):
                if move.from_warehouse != move.to_warehouse:
                    raise MoveValidationError(gettext(
                            'product_cost_warehouse'
                            '.msg_move_storage_location_same_warehouse',
                            from_=move.from_location.rec_name,
                            to=move.to_location.rec_name))

    def _do(self):
        cost_price, to_save = super()._do()
        if (self.company.cost_price_warehouse
                and self.from_location.type == 'storage'
                and self.to_location.type == 'storage'
                and self.from_warehouse != self.to_warehouse):
            if self.from_warehouse:
                cost_price = self._compute_product_cost_price('out')
            elif self.to_warehouse:
                cost_price = self._compute_product_cost_price(
                    'in', self.unit_price_company)
        return cost_price, to_save

    @property
    def _cost_price_pattern(self):
        pattern = super()._cost_price_pattern
        if self.company.cost_price_warehouse:
            pattern['warehouse'] = (
                self.warehouse.id if self.warehouse else None)
        return pattern

    def _cost_price_key(self):
        key = super()._cost_price_key()
        if self.company.cost_price_warehouse:
            key += (('warehouse',
                    (self.warehouse.id if self.warehouse else None)),
                )
        return key

    @classmethod
    def _cost_price_context(cls, moves):
        pool = Pool()
        Location = pool.get('stock.location')
        context = super()._cost_price_context(moves)
        if moves[0].company.cost_price_warehouse:
            warehouse = moves[0].warehouse
            locations = Location.search([
                    ('type', '=', 'storage'),
                    ['OR',
                        ('parent', 'child_of',
                            warehouse.id if warehouse else []),
                        ('cost_warehouse', '=',
                            warehouse.id if warehouse else None),
                        ],
                    ])
            context['locations'] = [l.id for l in locations]
        return context

    def get_fifo_move(self, quantity=0.0, date=None):
        warehouse = self.warehouse.id if self.warehouse else None
        with Transaction().set_context(warehouse=warehouse):
            return super().get_fifo_move(quantity=quantity, date=date)

    def _get_account_stock_move_type(self):
        type_ = super()._get_account_stock_move_type()
        if (self.company.cost_price_warehouse
                and self.from_location.type == 'storage'
                and self.to_location.type == 'storage'
                and self.from_warehouse != self.to_warehouse):
            if self.from_warehouse and not self.to_warehouse:
                type_ = 'out_warehouse'
            elif not self.from_warehouse and self.to_warehouse:
                type_ = 'in_warehouse'
        return type_


class ShipmentInternal(metaclass=PoolMeta):
    __name__ = 'stock.shipment.internal'

    @fields.depends('company')
    def on_change_with_transit_location(self, name=None):
        pool = Pool()
        Config = pool.get('stock.configuration')
        location = super().on_change_with_transit_location(name=name)
        if not location and self.company and self.company.cost_price_warehouse:
            location = Config(1).get_multivalue(
                'shipment_internal_transit', company=self.company)
        return location
